@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # Copyright (c) 2016, 2017 Red Hat, Inc.
 # Copyright (c) 2018 Nicira, Inc.
 #
@@ -21,67 +21,84 @@ import os
 import re
 import sys
 
-try:
-    import enchant
-
-    extra_keywords = ['ovs', 'vswitch', 'vswitchd', 'ovs-vswitchd', 'netdev',
-                      'selinux', 'ovs-ctl', 'dpctl', 'ofctl', 'openvswitch',
-                      'dpdk', 'hugepage', 'hugepages', 'pmd', 'upcall',
-                      'vhost', 'rx', 'tx', 'vhostuser', 'openflow', 'qsort',
-                      'rxq', 'txq', 'perf', 'stats', 'struct', 'int',
-                      'char', 'bool', 'upcalls', 'nicira', 'bitmask', 'ipv4',
-                      'ipv6', 'tcp', 'tcp4', 'tcpv4', 'udp', 'udp4', 'udpv4',
-                      'icmp', 'icmp4', 'icmpv6', 'vlan', 'vxlan', 'cksum',
-                      'csum', 'checksum', 'ofproto', 'numa', 'mempool',
-                      'mempools', 'mbuf', 'mbufs', 'hmap', 'cmap', 'smap',
-                      'dhcpv4', 'dhcp', 'dhcpv6', 'opts', 'metadata',
-                      'geneve', 'mutex', 'netdev', 'netdevs', 'subtable',
-                      'virtio', 'qos', 'policer', 'datapath', 'tunctl',
-                      'attr', 'ethernet', 'ether', 'defrag', 'defragment',
-                      'loopback', 'sflow', 'acl', 'initializer', 'recirc',
-                      'xlated', 'unclosed', 'netlink', 'msec', 'usec',
-                      'nsec', 'ms', 'us', 'ns', 'kilobits', 'kbps',
-                      'kilobytes', 'megabytes', 'mbps', 'gigabytes', 'gbps',
-                      'megabits', 'gigabits', 'pkts', 'tuple', 'miniflow',
-                      'megaflow', 'conntrack', 'vlans', 'vxlans', 'arg',
-                      'tpid', 'xbundle', 'xbundles', 'mbundle', 'mbundles',
-                      'netflow', 'localnet', 'odp', 'pre', 'dst', 'dest',
-                      'src', 'ethertype', 'cvlan', 'ips', 'msg', 'msgs',
-                      'liveness', 'userspace', 'eventmask', 'datapaths',
-                      'slowpath', 'fastpath', 'multicast', 'unicast',
-                      'revalidation', 'namespace', 'qdisc', 'uuid', 'ofport',
-                      'subnet', 'revalidation', 'revalidator', 'revalidate',
-                      'l2', 'l3', 'l4', 'openssl', 'mtu', 'ifindex', 'enum',
-                      'enums', 'http', 'https', 'num', 'vconn', 'vconns',
-                      'conn', 'nat', 'memset', 'memcmp', 'strcmp',
-                      'strcasecmp', 'tc', 'ufid', 'api', 'ofpbuf', 'ofpbufs',
-                      'hashmaps', 'hashmap', 'deref', 'dereference', 'hw',
-                      'prio', 'sendmmsg', 'sendmsg', 'malloc', 'free', 'alloc',
-                      'pid', 'ppid', 'pgid', 'uid', 'gid', 'sid', 'utime',
-                      'stime', 'cutime', 'cstime', 'vsize', 'rss', 'rsslim',
-                      'whcan', 'gtime', 'eip', 'rip', 'cgtime', 'dbg', 'gw',
-                      'sbrec', 'bfd', 'sizeof', 'pmds', 'nic', 'nics', 'hwol',
-                      'encap', 'decap', 'tlv', 'tlvs', 'decapsulation', 'fd',
-                      'cacheline', 'xlate', 'skiplist', 'idl', 'comparator',
-                      'natting', 'alg', 'pasv', 'epasv', 'wildcard', 'nated',
-                      'amd64', 'x86_64', 'recirculation']
-
-    spell_check_dict = enchant.Dict("en_US")
-    for kw in extra_keywords:
-        spell_check_dict.add(kw)
-
-    no_spellcheck = False
-except:
-    no_spellcheck = True
-
+RETURN_CHECK_INITIAL_STATE = 0
+RETURN_CHECK_STATE_WITH_RETURN = 1
+RETURN_CHECK_AWAITING_BRACE = 2
+EXIT_FAILURE = 1
 __errors = 0
 __warnings = 0
+empty_return_check_state = 0
 print_file_name = None
 checking_file = False
 total_line = 0
 colors = False
-spellcheck_comments = False
+spellcheck = False
 quiet = False
+spell_check_dict = None
+
+
+def open_spell_check_dict():
+    import enchant
+
+    try:
+        extra_keywords = ['ovs', 'vswitch', 'vswitchd', 'ovs-vswitchd',
+                          'netdev', 'selinux', 'ovs-ctl', 'dpctl', 'ofctl',
+                          'openvswitch', 'dpdk', 'hugepage', 'hugepages',
+                          'pmd', 'upcall', 'vhost', 'rx', 'tx', 'vhostuser',
+                          'openflow', 'qsort', 'rxq', 'txq', 'perf', 'stats',
+                          'struct', 'int', 'char', 'bool', 'upcalls', 'nicira',
+                          'bitmask', 'ipv4', 'ipv6', 'tcp', 'tcp4', 'tcpv4',
+                          'udp', 'udp4', 'udpv4', 'icmp', 'icmp4', 'icmpv6',
+                          'vlan', 'vxlan', 'cksum', 'csum', 'checksum',
+                          'ofproto', 'numa', 'mempool', 'mempools', 'mbuf',
+                          'mbufs', 'hmap', 'cmap', 'smap', 'dhcpv4', 'dhcp',
+                          'dhcpv6', 'opts', 'metadata', 'geneve', 'mutex',
+                          'netdev', 'netdevs', 'subtable', 'virtio', 'qos',
+                          'policer', 'datapath', 'tunctl', 'attr', 'ethernet',
+                          'ether', 'defrag', 'defragment', 'loopback', 'sflow',
+                          'acl', 'initializer', 'recirc', 'xlated', 'unclosed',
+                          'netlink', 'msec', 'usec', 'nsec', 'ms', 'us', 'ns',
+                          'kilobits', 'kbps', 'kilobytes', 'megabytes', 'mbps',
+                          'gigabytes', 'gbps', 'megabits', 'gigabits', 'pkts',
+                          'tuple', 'miniflow', 'megaflow', 'conntrack',
+                          'vlans', 'vxlans', 'arg', 'tpid', 'xbundle',
+                          'xbundles', 'mbundle', 'mbundles', 'netflow',
+                          'localnet', 'odp', 'pre', 'dst', 'dest', 'src',
+                          'ethertype', 'cvlan', 'ips', 'msg', 'msgs',
+                          'liveness', 'userspace', 'eventmask', 'datapaths',
+                          'slowpath', 'fastpath', 'multicast', 'unicast',
+                          'revalidation', 'namespace', 'qdisc', 'uuid',
+                          'ofport', 'subnet', 'revalidation', 'revalidator',
+                          'revalidate', 'l2', 'l3', 'l4', 'openssl', 'mtu',
+                          'ifindex', 'enum', 'enums', 'http', 'https', 'num',
+                          'vconn', 'vconns', 'conn', 'nat', 'memset', 'memcmp',
+                          'strcmp', 'strcasecmp', 'tc', 'ufid', 'api',
+                          'ofpbuf', 'ofpbufs', 'hashmaps', 'hashmap', 'deref',
+                          'dereference', 'hw', 'prio', 'sendmmsg', 'sendmsg',
+                          'malloc', 'free', 'alloc', 'pid', 'ppid', 'pgid',
+                          'uid', 'gid', 'sid', 'utime', 'stime', 'cutime',
+                          'cstime', 'vsize', 'rss', 'rsslim', 'whcan', 'gtime',
+                          'eip', 'rip', 'cgtime', 'dbg', 'gw', 'sbrec', 'bfd',
+                          'sizeof', 'pmds', 'nic', 'nics', 'hwol', 'encap',
+                          'decap', 'tlv', 'tlvs', 'decapsulation', 'fd',
+                          'cacheline', 'xlate', 'skiplist', 'idl',
+                          'comparator', 'natting', 'alg', 'pasv', 'epasv',
+                          'wildcard', 'nated', 'amd64', 'x86_64',
+                          'recirculation', 'linux', 'afxdp', 'promisc', 'goto',
+                          'misconfigured', 'misconfiguration', 'checkpatch',
+                          'debian', 'travis', 'cirrus', 'appveyor', 'faq',
+                          'erspan', 'const', 'hotplug', 'addresssanitizer',
+                          'ovsdb', 'dpif', 'veth', 'rhel', 'jsonrpc', 'json',
+                          'syscall', 'lacp', 'ipf', 'skb', 'valgrind']
+
+        global spell_check_dict
+        spell_check_dict = enchant.Dict("en_US")
+        for kw in extra_keywords:
+            spell_check_dict.add(kw)
+
+        return True
+    except:
+        return False
 
 
 def get_color_end():
@@ -131,7 +148,7 @@ def reset_counters():
 # something in parentheses (usually an expression) then a left curly brace.
 #
 # 'do' almost qualifies but it's also used as "do { ... } while (...);".
-__parenthesized_constructs = 'if|for|while|switch|[_A-Z]+FOR_EACH[_A-Z]*'
+__parenthesized_constructs = 'if|for|while|switch|[_A-Z]+FOR_*EACH[_A-Z0-9]*'
 
 __regex_added_line = re.compile(r'^\+{1,2}[^\+][\w\W]*')
 __regex_subtracted_line = re.compile(r'^\-{1,2}[^\-][\w\W]*')
@@ -150,17 +167,23 @@ __regex_is_for_if_single_line_bracket = \
 __regex_ends_with_bracket = \
     re.compile(r'[^\s]\) {(\s+/\*[\s\Sa-zA-Z0-9\.,\?\*/+-]*)?$')
 __regex_ptr_declaration_missing_whitespace = re.compile(r'[a-zA-Z0-9]\*[^*]')
+__regex_cast_missing_whitespace = re.compile(r'\)[a-zA-Z0-9]')
 __regex_is_comment_line = re.compile(r'^\s*(/\*|\*\s)')
 __regex_has_comment = re.compile(r'.*(/\*|\*\s)')
+__regex_has_c99_comment = re.compile(r'.*//.*$')
 __regex_trailing_operator = re.compile(r'^[^ ]* [^ ]*[?:]$')
 __regex_conditional_else_bracing = re.compile(r'^\s*else\s*{?$')
 __regex_conditional_else_bracing2 = re.compile(r'^\s*}\selse\s*$')
 __regex_has_xxx_mark = re.compile(r'.*xxx.*', re.IGNORECASE)
 __regex_added_doc_rst = re.compile(
                     r'\ndiff .*Documentation/.*rst\nnew file mode')
+__regex_empty_return = re.compile(r'\s*return;')
+__regex_if_macros = re.compile(r'^ +(%s) \([\S]([\s\S]+[\S])*\) { +\\' %
+                               __parenthesized_constructs)
 
 skip_leading_whitespace_check = False
 skip_trailing_whitespace_check = False
+skip_gerrit_change_id_check = False
 skip_block_whitespace_check = False
 skip_signoff_check = False
 
@@ -168,13 +191,13 @@ skip_signoff_check = False
 # name, as they may have legitimate reasons to have longer lines.
 #
 # Python isn't checked as flake8 performs these checks during build.
-line_length_blacklist = re.compile(
+line_length_ignore_list = re.compile(
     r'\.(am|at|etc|in|m4|mk|patch|py)$|debian/rules')
 
 # Don't enforce a requirement that leading whitespace be all spaces on
 # files that include these characters in their name, since these kinds
 # of files need lines with leading tabs.
-leading_whitespace_blacklist = re.compile(r'\.(mk|am|at)$|debian/rules')
+leading_whitespace_ignore_list = re.compile(r'\.(mk|am|at)$|debian/rules')
 
 
 def is_subtracted_line(line):
@@ -247,7 +270,9 @@ def if_and_for_end_with_bracket_check(line):
     if __regex_is_for_if_single_line_bracket.search(line) is not None:
         if not balanced_parens(line):
             return True
-        if __regex_ends_with_bracket.search(line) is None:
+
+        if __regex_ends_with_bracket.search(line) is None and \
+           __regex_if_macros.match(line) is None:
             return False
     if __regex_conditional_else_bracing.match(line) is not None:
         return False
@@ -262,6 +287,12 @@ def pointer_whitespace_check(line):
     return __regex_ptr_declaration_missing_whitespace.search(line) is not None
 
 
+def cast_whitespace_check(line):
+    """Return TRUE if there is no space between the '()' used in a cast and
+       the expression whose type is cast, i.e.: '(void *)foo'"""
+    return __regex_cast_missing_whitespace.search(line) is not None
+
+
 def line_length_check(line):
     """Return TRUE if the line length is too long"""
     if len(line) > 79:
@@ -282,6 +313,11 @@ def has_comment(line):
     return __regex_has_comment.match(line) is not None
 
 
+def has_c99_comment(line):
+    """Returns TRUE if the current line contains C99 style comment (//)."""
+    return __regex_has_c99_comment.match(line) is not None
+
+
 def trailing_operator(line):
     """Returns TRUE if the current line ends with an operatorsuch as ? or :"""
     return __regex_trailing_operator.match(line) is not None
@@ -349,15 +385,19 @@ def filter_comments(current_line, keep=False):
     return sanitized_line
 
 
-def check_comment_spelling(line):
-    if no_spellcheck or not spellcheck_comments:
+def check_spelling(line, comment):
+    if not spell_check_dict or not spellcheck:
         return False
 
-    comment_words = filter_comments(line, True).replace(':', ' ').split(' ')
-    for word in comment_words:
+    words = filter_comments(line, True) if comment else line
+    words = words.replace(':', ' ').split(' ')
+
+    for word in words:
         skip = False
         strword = re.subn(r'\W+', '', word)[0].replace(',', '')
-        if len(strword) and not spell_check_dict.check(strword.lower()):
+        if (len(strword)
+                and not spell_check_dict.check(strword.lower())
+                and not spell_check_dict.check(word.lower())):
             if any([check_char in word
                     for check_char in ['=', '(', '-', '_', '/', '\'']]):
                 skip = True
@@ -371,8 +411,8 @@ def check_comment_spelling(line):
                                      strword[1:].islower()):
                 skip = True
 
-            # skip words that start with numbers
-            if strword.startswith(tuple('0123456789')):
+            # skip words containing numbers
+            if any(check_char.isdigit() for check_char in strword):
                 skip = True
 
             if not skip:
@@ -454,6 +494,33 @@ def check_new_docs_index(text):
     return __check_new_docs(text, 'rst')
 
 
+def empty_return_with_brace(line):
+    """Returns TRUE if a function contains a return; followed
+       by one or more line feeds and terminates with a '}'
+       at start of line"""
+
+    def empty_return(line):
+        """Returns TRUE if a function has a 'return;'"""
+        return __regex_empty_return.match(line) is not None
+
+    global empty_return_check_state
+    if empty_return_check_state == RETURN_CHECK_INITIAL_STATE \
+       and empty_return(line):
+        empty_return_check_state = RETURN_CHECK_STATE_WITH_RETURN
+    elif empty_return_check_state == RETURN_CHECK_STATE_WITH_RETURN \
+         and (re.match(r'^}$', line) or len(line) == 0):
+        if re.match('^}$', line):
+            empty_return_check_state = RETURN_CHECK_AWAITING_BRACE
+    else:
+        empty_return_check_state = RETURN_CHECK_INITIAL_STATE
+
+    if empty_return_check_state == RETURN_CHECK_AWAITING_BRACE:
+        empty_return_check_state = RETURN_CHECK_INITIAL_STATE
+        return True
+
+    return False
+
+
 file_checks = [
         {'regex': __regex_added_doc_rst,
          'check': check_new_docs_index},
@@ -463,11 +530,11 @@ file_checks = [
 
 checks = [
     {'regex': None,
-     'match_name': lambda x: not line_length_blacklist.search(x),
+     'match_name': lambda x: not line_length_ignore_list.search(x),
      'check': lambda x: line_length_check(x)},
 
     {'regex': None,
-     'match_name': lambda x: not leading_whitespace_blacklist.search(x),
+     'match_name': lambda x: not leading_whitespace_ignore_list.search(x),
      'check': lambda x: not leading_whitespace_is_spaces(x),
      'print': lambda: print_warning("Line has non-spaces leading whitespace")},
 
@@ -475,36 +542,54 @@ checks = [
      'check': lambda x: trailing_whitespace_or_crlf(x),
      'print': lambda: print_warning("Line has trailing whitespace")},
 
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: not is_comment_line(x),
      'check': lambda x: not if_and_for_whitespace_checks(x),
      'print': lambda: print_error("Improper whitespace around control block")},
 
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: not is_comment_line(x),
      'check': lambda x: not if_and_for_end_with_bracket_check(x),
      'print': lambda: print_error("Inappropriate bracing around statement")},
 
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: not is_comment_line(x),
      'check': lambda x: pointer_whitespace_check(x),
      'print':
      lambda: print_error("Inappropriate spacing in pointer declaration")},
 
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
+     'prereq': lambda x: not is_comment_line(x),
+     'check': lambda x: cast_whitespace_check(x),
+     'print':
+     lambda: print_error("Inappropriate spacing around cast")},
+
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: not is_comment_line(x),
      'check': lambda x: trailing_operator(x),
      'print':
      lambda: print_error("Line has '?' or ':' operator at end of line")},
 
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: has_comment(x),
      'check': lambda x: has_xxx_mark(x),
      'print': lambda: print_warning("Comment with 'xxx' marker")},
 
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
+     'prereq': lambda x: not is_comment_line(x),
+     'check': lambda x: has_c99_comment(x),
+     'print': lambda: print_error("C99 style comment")},
+
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: has_comment(x),
-     'check': lambda x: check_comment_spelling(x)},
+     'check': lambda x: check_spelling(x, True)},
+
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
+     'check': lambda x: empty_return_with_brace(x),
+     'interim_line': True,
+     'print':
+     lambda: print_warning("Empty return followed by brace, consider omitting")
+     },
 ]
 
 
@@ -534,7 +619,7 @@ std_functions = [
         ('error', 'Use ovs_error() in place of error()'),
 ]
 checks += [
-    {'regex': '(\.c|\.h)(\.in)?$',
+    {'regex': r'(\.c|\.h)(\.in)?$',
      'match_name': None,
      'prereq': lambda x: not is_comment_line(x),
      'check': regex_function_factory(function_name),
@@ -551,11 +636,16 @@ infix_operators = \
     [re.escape(op) for op in ['%', '<<', '>>', '<=', '>=', '==', '!=',
             '^', '|', '&&', '||', '?:', '=', '+=', '-=', '*=', '/=', '%=',
             '&=', '^=', '|=', '<<=', '>>=']] \
-    + ['[^<" ]<[^=" ]', '[^->" ]>[^=" ]', '[^ !()/"]\*[^/]', '[^ !&()"]&',
-       '[^" +(]\+[^"+;]', '[^" -(]-[^"->;]', '[^" <>=!^|+\-*/%&]=[^"=]',
-       '[^* ]/[^* ]']
+    + [r'[^<" ]<[^=" ]',
+       r'[^\->" ]>[^=" ]',
+       r'[^ !()/"]\*[^/]',
+       r'[^ !&()"]&',
+       r'[^" +(]\+[^"+;]',
+       r'[^" \-(]\-[^"\->;]',
+       r'[^" <>=!^|+\-*/%&]=[^"=]',
+       r'[^* ]/[^* ]']
 checks += [
-    {'regex': '(\.c|\.h)(\.in)?$', 'match_name': None,
+    {'regex': r'(\.c|\.h)(\.in)?$', 'match_name': None,
      'prereq': lambda x: not is_comment_line(x),
      'check': regex_operator_factory(operator),
      'print': lambda: print_warning("Line lacks whitespace around operator")}
@@ -599,6 +689,29 @@ def run_checks(current_file, line, lineno):
         print("%s\n" % line)
 
 
+def interim_line_check(current_file, line, lineno):
+    """Runs the various checks for the particular interim line.  This will
+       take filename into account, and will check for the 'interim_line'
+       key before running the check."""
+    global checking_file, total_line
+    print_line = False
+    for check in get_file_type_checks(current_file):
+        if 'prereq' in check and not check['prereq'](line):
+            continue
+        if 'interim_line' in check and check['interim_line']:
+            if check['check'](line):
+                if 'print' in check:
+                    check['print']()
+                    print_line = True
+
+    if print_line:
+        if checking_file:
+            print("%s:%d:" % (current_file, lineno))
+        else:
+            print("#%d FILE: %s:%d:" % (total_line, current_file, lineno))
+        print("%s\n" % line)
+
+
 def run_file_checks(text):
     """Runs the various checks for the text."""
     for check in file_checks:
@@ -606,8 +719,9 @@ def run_file_checks(text):
             check['check'](text)
 
 
-def ovs_checkpatch_parse(text, filename):
-    global print_file_name, total_line, checking_file
+def ovs_checkpatch_parse(text, filename, author=None, committer=None):
+    global print_file_name, total_line, checking_file, \
+        empty_return_check_state
 
     PARSE_STATE_HEADING = 0
     PARSE_STATE_DIFF_HEADER = 1
@@ -620,9 +734,11 @@ def ovs_checkpatch_parse(text, filename):
     current_file = filename if checking_file else ''
     previous_file = ''
     seppatch = re.compile(r'^---([\w]*| \S+)$')
-    hunks = re.compile('^(---|\+\+\+) (\S+)')
+    hunks = re.compile(r'^(---|\+\+\+) (\S+)')
     hunk_differences = re.compile(
         r'^@@ ([0-9-+]+),([0-9-+]+) ([0-9-+]+),([0-9-+]+) @@')
+    is_author = re.compile(r'^(Author|From): (.*)$', re.I | re.M | re.S)
+    is_committer = re.compile(r'^(Commit: )(.*)$', re.I | re.M | re.S)
     is_signature = re.compile(r'^(Signed-off-by: )(.*)$',
                               re.I | re.M | re.S)
     is_co_author = re.compile(r'^(Co-authored-by: )(.*)$',
@@ -632,7 +748,7 @@ def ovs_checkpatch_parse(text, filename):
 
     reset_counters()
 
-    for line in text.split('\n'):
+    for line in text.splitlines():
         if current_file != previous_file:
             previous_file = current_file
 
@@ -655,23 +771,71 @@ def ovs_checkpatch_parse(text, filename):
             if seppatch.match(line):
                 parse = PARSE_STATE_DIFF_HEADER
                 if not skip_signoff_check:
-                    if len(signatures) == 0:
-                        print_error("No signatures found.")
-                    elif len(signatures) != 1 + len(co_authors):
-                        print_error("Too many signoffs; "
-                                    "are you missing Co-authored-by lines?")
-                    if not set(co_authors) <= set(signatures):
-                        print_error("Co-authored-by/Signed-off-by corruption")
+
+                    # Check that the patch has an author, that the
+                    # author is not among the co-authors, and that the
+                    # co-authors are unique.
+                    if not author:
+                        print_error("Patch lacks author.")
+                        continue
+                    if " via " in author or "@openvswitch.org" in author:
+                        print_error("Author should not be mailing list.")
+                        continue
+                    if author in co_authors:
+                        print_error("Author should not be also be co-author.")
+                        continue
+                    if len(set(co_authors)) != len(co_authors):
+                        print_error("Duplicate co-author.")
+
+                    # Check that the author, all co-authors, and the
+                    # committer (if any) signed off.
+                    if author not in signatures:
+                        print_error("Author %s needs to sign off." % author)
+                    for ca in co_authors:
+                        if ca not in signatures:
+                            print_error("Co-author %s needs to sign off." % ca)
+                            break
+                    if (committer
+                        and author != committer
+                        and committer not in signatures):
+                        print_error("Committer %s needs to sign off."
+                                    % committer)
+
+                    # Check for signatures that we do not expect.
+                    # This is only a warning because there can be,
+                    # rarely, a signature chain.
+                    #
+                    # If we don't have a known committer, and there is
+                    # a single extra sign-off, then do not warn
+                    # because that extra sign-off is probably the
+                    # committer.
+                    extra_sigs = [x for x in signatures
+                                  if x not in co_authors
+                                  and x != author
+                                  and x != committer]
+                    if len(extra_sigs) > 1 or (committer and extra_sigs):
+                        print_warning("Unexpected sign-offs from developers "
+                                      "who are not authors or co-authors or "
+                                      "committers: %s"
+                                      % ", ".join(extra_sigs))
+            elif is_committer.match(line):
+                committer = is_committer.match(line).group(2)
+            elif is_author.match(line):
+                author = is_author.match(line).group(2)
             elif is_signature.match(line):
                 m = is_signature.match(line)
                 signatures.append(m.group(2))
             elif is_co_author.match(line):
                 m = is_co_author.match(line)
                 co_authors.append(m.group(2))
-            elif is_gerrit_change_id.match(line):
+            elif (is_gerrit_change_id.match(line) and
+                  not skip_gerrit_change_id_check):
                 print_error(
                     "Remove Gerrit Change-Id's before submitting upstream.")
                 print("%d: %s\n" % (lineno, line))
+            elif spellcheck:
+                check_spelling(line, False)
+
         elif parse == PARSE_STATE_CHANGE_BODY:
             newfile = hunks.match(line)
             if newfile:
@@ -680,28 +844,40 @@ def ovs_checkpatch_parse(text, filename):
                 continue
             reset_line_number = hunk_differences.match(line)
             if reset_line_number:
+                empty_return_check_state = RETURN_CHECK_INITIAL_STATE
                 lineno = int(reset_line_number.group(3))
                 if lineno < 0:
                     lineno = -1 * lineno
                 lineno -= 1
+
             if is_subtracted_line(line):
                 lineno -= 1
-            if not is_added_line(line):
                 continue
 
             cmp_line = added_line(line)
 
+            if not is_added_line(line):
+                interim_line_check(current_file, cmp_line, lineno)
+                continue
+
             # Skip files which have /datapath in them, since they are
             # linux or windows coding standards
             if current_file.startswith('datapath'):
                 continue
             if current_file.startswith('include/linux'):
                 continue
+            # "sparse" includes could be copy-pasted from different sources
+            # like DPDK or Linux and could contain workarounds not suitable
+            # for a common style.
+            if current_file.startswith('include/sparse'):
+                continue
+            if current_file.startswith('utilities/bugtool'):
+                continue
             run_checks(current_file, cmp_line, lineno)
 
     run_file_checks(text)
     if __errors or __warnings:
-        return -1
+        return EXIT_FAILURE
     return 0
 
 
@@ -722,15 +898,17 @@ Check options:
 -l|--skip-leading-whitespace   Skips the leading whitespace test
 -q|--quiet                     Only print error and warning information
 -s|--skip-signoff-lines        Tolerate missing Signed-off-by line
--S|--spellcheck-comments       Check C comments for possible spelling mistakes
--t|--skip-trailing-whitespace  Skips the trailing whitespace test"""
+-S|--spellcheck                Check C comments and commit-message for possible
+                               spelling mistakes
+-t|--skip-trailing-whitespace  Skips the trailing whitespace test
+   --skip-gerrit-change-id     Skips the gerrit change id test"""
           % sys.argv[0])
 
 
-def ovs_checkpatch_print_result(result):
+def ovs_checkpatch_print_result():
     global quiet, __warnings, __errors, total_line
 
-    if result < 0:
+    if __errors or __warnings:
         print("Lines checked: %d, Warnings: %d, Errors: %d\n" %
               (total_line, __warnings, __errors))
     elif not quiet:
@@ -747,8 +925,10 @@ def ovs_checkpatch_file(filename):
     for part in mail.walk():
         if part.get_content_maintype() == 'multipart':
             continue
-    result = ovs_checkpatch_parse(part.get_payload(decode=False), filename)
-    ovs_checkpatch_print_result(result)
+    result = ovs_checkpatch_parse(part.get_payload(decode=False), filename,
+                                  mail.get('Author', mail['From']),
+                                  mail['Commit'])
+    ovs_checkpatch_print_result()
     return result
 
 
@@ -778,11 +958,12 @@ if __name__ == '__main__':
                                        "skip-leading-whitespace",
                                        "skip-signoff-lines",
                                        "skip-trailing-whitespace",
-                                       "spellcheck-comments",
+                                       "skip-gerrit-change-id",
+                                       "spellcheck",
                                        "quiet"])
     except:
         print("Unknown option encountered. Please rerun with -h for help.")
-        sys.exit(-1)
+        sys.exit(EXIT_FAILURE)
 
     for o, a in optlist:
         if o in ("-h", "--help"):
@@ -796,19 +977,21 @@ if __name__ == '__main__':
             skip_signoff_check = True
         elif o in ("-t", "--skip-trailing-whitespace"):
             skip_trailing_whitespace_check = True
+        elif o in ("--skip-gerrit-change-id"):
+            skip_gerrit_change_id_check = True
         elif o in ("-f", "--check-file"):
             checking_file = True
-        elif o in ("-S", "--spellcheck-comments"):
-            if no_spellcheck:
-                print("WARNING: The enchant library isn't availble.")
+        elif o in ("-S", "--spellcheck"):
+            if not open_spell_check_dict():
+                print("WARNING: The enchant library isn't available.")
                 print("         Please install python enchant.")
             else:
-                spellcheck_comments = True
+                spellcheck = True
         elif o in ("-q", "--quiet"):
             quiet = True
         else:
             print("Unknown option '%s'" % o)
-            sys.exit(-1)
+            sys.exit(EXIT_FAILURE)
 
     if sys.stdout.isatty():
         colors = True
@@ -822,24 +1005,29 @@ if __name__ == '__main__':
 
         for i in reversed(range(0, n_patches)):
             revision, name = commits[i].split(" ", 1)
-            f = os.popen('git format-patch -1 --stdout %s' % revision, 'r')
+            f = os.popen('''git format-patch -1 --stdout --pretty=format:"\
+Author: %an <%ae>
+Commit: %cn <%ce>
+Subject: %s
+
+%b" ''' + revision, 'r')
             patch = f.read()
             f.close()
 
             if not quiet:
                 print('== Checking %s ("%s") ==' % (revision[0:12], name))
             result = ovs_checkpatch_parse(patch, revision)
-            ovs_checkpatch_print_result(result)
+            ovs_checkpatch_print_result()
             if result:
-                status = -1
+                status = EXIT_FAILURE
         sys.exit(status)
 
     if not args:
         if sys.stdin.isatty():
             usage()
-            sys.exit(-1)
+            sys.exit(EXIT_FAILURE)
         result = ovs_checkpatch_parse(sys.stdin.read(), '-')
-        ovs_checkpatch_print_result(result)
+        ovs_checkpatch_print_result()
         sys.exit(result)
 
     status = 0
@@ -848,5 +1036,5 @@ if __name__ == '__main__':
             print('== Checking "%s" ==' % filename)
         result = ovs_checkpatch_file(filename)
         if result:
-            status = -1
+            status = EXIT_FAILURE
     sys.exit(status)
