@@ -30,6 +30,7 @@
 #    - ovs_dump_netdev_provider
 #    - ovs_dump_ovs_list <struct ovs_list *> {[<structure>] [<member>] {dump}]}
 #    - ovs_dump_simap <struct simap *>
+#    - ovs_dump_smap <struct smap *>
 #    - ovs_dump_udpif_keys {<udpif_name>|<udpif_address>} {short}
 #    - ovs_show_fdb {[<bridge_name>] {dbg} {hash}}
 #    - ovs_show_upcall {dbg}
@@ -54,7 +55,9 @@
 #    ...
 #    ...
 #
+from __future__ import print_function
 import gdb
+import sys
 import uuid
 
 
@@ -121,7 +124,7 @@ def get_global_variable(name):
     var = gdb.lookup_symbol(name)[0]
     if var is None or not var.is_variable:
         print("Can't find {} global variable, are you sure "
-              "your debugging OVS?".format(name))
+              "you are debugging OVS?".format(name))
         return None
     return gdb.parse_and_eval(name)
 
@@ -131,22 +134,53 @@ def get_time_msec():
     # to get a decent time time_now() value. For now we take the global
     # "coverage_run_time" value, which is the current time + max 5 seconds
     # (COVERAGE_RUN_INTERVAL)
-    return long(get_global_variable("coverage_run_time")), -5000
+    return int(get_global_variable("coverage_run_time")), -5000
 
 
 def get_time_now():
     # See get_time_msec() above
-    return long(get_global_variable("coverage_run_time"))/1000, -5
+    return int(get_global_variable("coverage_run_time"))/1000, -5
 
 
 def eth_addr_to_string(eth_addr):
     return "{:02x}:{:02x}:{:02x}:{:02x}:{:02x}:{:02x}".format(
-        long(eth_addr['ea'][0]),
-        long(eth_addr['ea'][1]),
-        long(eth_addr['ea'][2]),
-        long(eth_addr['ea'][3]),
-        long(eth_addr['ea'][4]),
-        long(eth_addr['ea'][5]))
+        int(eth_addr['ea'][0]),
+        int(eth_addr['ea'][1]),
+        int(eth_addr['ea'][2]),
+        int(eth_addr['ea'][3]),
+        int(eth_addr['ea'][4]),
+        int(eth_addr['ea'][5]))
+
+
+#
+# Simple class to print a spinner on the console
+#
+class ProgressIndicator(object):
+    def __init__(self, message=None):
+        self.spinner = "/-\|"
+        self.spinner_index = 0
+        self.message = message
+
+        if self.message is not None:
+            print(self.message, end='')
+
+    def update(self):
+        print("{}\b".format(self.spinner[self.spinner_index]), end='')
+        sys.stdout.flush()
+        self.spinner_index += 1
+        if self.spinner_index >= len(self.spinner):
+            self.spinner_index = 0
+
+    def pause(self):
+        print("\r\033[K", end='')
+
+    def resume(self):
+        if self.message is not None:
+            print(self.message, end='')
+        self.update()
+
+    def done(self):
+        self.pause()
 
 
 #
@@ -191,7 +225,7 @@ class ForEachCMAP(object):
 
         raise StopIteration
 
-    def next(self):
+    def __next__(self):
         ipml = self.cmap['impl']['p']
         if ipml['n'] == 0:
             raise StopIteration
@@ -205,6 +239,9 @@ class ForEachCMAP(object):
                             gdb.lookup_type(self.typeobj).pointer(),
                             self.member)
 
+    def next(self):
+        return self.__next__()
+
 
 #
 # Class that will provide an iterator over an OVS hmap.
@@ -228,7 +265,7 @@ class ForEachHMAP(object):
 
         raise StopIteration
 
-    def next(self):
+    def __next__(self):
         #
         # In the real implementation the n values is never checked,
         # however when debugging we do, as we might try to access
@@ -252,6 +289,9 @@ class ForEachHMAP(object):
                             gdb.lookup_type(self.typeobj).pointer(),
                             self.member)
 
+    def next(self):
+        return self.__next__()
+
 
 #
 # Class that will provide an iterator over an Netlink attributes
@@ -259,7 +299,7 @@ class ForEachHMAP(object):
 class ForEachNL():
     def __init__(self, nlattrs, nlattrs_len):
         self.attr = nlattrs.cast(gdb.lookup_type('struct nlattr').pointer())
-        self.attr_len = long(nlattrs_len)
+        self.attr_len = int(nlattrs_len)
 
     def __iter__(self):
         return self
@@ -267,7 +307,7 @@ class ForEachNL():
     def round_up(self, val, round_to):
         return int(val) + (round_to - int(val)) % round_to
 
-    def next(self):
+    def __next__(self):
         if self.attr is None or \
            self.attr_len < 4 or self.attr['nla_len'] < 4 or  \
            self.attr['nla_len'] > self.attr_len:
@@ -285,6 +325,9 @@ class ForEachNL():
 
         return attr
 
+    def next(self):
+        return self.__next__()
+
 
 #
 # Class that will provide an iterator over an OVS shash.
@@ -297,14 +340,17 @@ class ForEachSHASH(ForEachHMAP):
         super(ForEachSHASH, self).__init__(shash['map'],
                                            "struct shash_node", "node")
 
-    def next(self):
-        node = super(ForEachSHASH, self).next()
+    def __next__(self):
+        node = super(ForEachSHASH, self).__next__()
 
         if self.data_typeobj is None:
             return node
 
         return node['data'].cast(gdb.lookup_type(self.data_typeobj).pointer())
 
+    def next(self):
+        return self.__next__()
+
 
 #
 # Class that will provide an iterator over an OVS simap.
@@ -314,10 +360,29 @@ class ForEachSIMAP(ForEachHMAP):
         super(ForEachSIMAP, self).__init__(shash['map'],
                                            "struct simap_node", "node")
 
-    def next(self):
-        node = super(ForEachSIMAP, self).next()
+    def __next__(self):
+        node = super(ForEachSIMAP, self).__next__()
         return node['name'], node['data']
 
+    def next(self):
+        return self.__next__()
+
+
+#
+# Class that will provide an iterator over an OVS smap.
+#
+class ForEachSMAP(ForEachHMAP):
+    def __init__(self, shash):
+        super(ForEachSMAP, self).__init__(shash['map'],
+                                          "struct smap_node", "node")
+
+    def __next__(self):
+        node = super(ForEachSMAP, self).__next__()
+        return node['key'], node['value']
+
+    def next(self):
+        return self.__next__()
+
 
 #
 # Class that will provide an iterator over an OVS list.
@@ -332,7 +397,7 @@ class ForEachLIST():
     def __iter__(self):
         return self
 
-    def next(self):
+    def __next__(self):
         if self.list.address == self.node['next']:
             raise StopIteration
 
@@ -345,6 +410,42 @@ class ForEachLIST():
                             gdb.lookup_type(self.typeobj).pointer(),
                             self.member)
 
+    def next(self):
+        return self.__next__()
+
+
+#
+# Class that will provide an iterator over an OFPACTS.
+#
+class ForEachOFPACTS():
+    def __init__(self, ofpacts, ofpacts_len):
+        self.ofpact = ofpacts.cast(gdb.lookup_type('struct ofpact').pointer())
+        self.length = int(ofpacts_len)
+
+    def __round_up(self, val, round_to):
+        return int(val) + (round_to - int(val)) % round_to
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        if self.ofpact is None or self.length <= 0:
+            raise StopIteration
+
+        ofpact = self.ofpact
+        length = self.__round_up(ofpact['len'], 8)
+
+        self.length -= length
+        self.ofpact = self.ofpact.cast(
+            gdb.lookup_type('void').pointer()) + length
+        self.ofpact = self.ofpact.cast(
+            gdb.lookup_type('struct ofpact').pointer())
+
+        return ofpact
+
+    def next(self):
+        return self.__next__()
+
 
 #
 # Implements the GDB "ovs_dump_bridges" command
@@ -410,7 +511,7 @@ class CmdDumpBridgePorts(gdb.Command):
     def display_single_port(port, indent=0):
         indent = " " * indent
         port = port.cast(gdb.lookup_type('struct port').pointer())
-        print("{}(struct port *) {}: name = {}, brige = (struct bridge *) {}".
+        print("{}(struct port *) {}: name = {}, bridge = (struct bridge *) {}".
               format(indent, port, port['name'].string(),
                      port['bridge']))
 
@@ -485,7 +586,7 @@ class CmdDumpDpNetdevPollThreads(gdb.Command):
     @staticmethod
     def display_single_poll_thread(pmd_thread, indent=0):
         indent = " " * indent
-        print("{}(struct dp_netdev_pmd_thread *) {}: core_id = {:s}, "
+        print("{}(struct dp_netdev_pmd_thread *) {}: core_id = {}, "
               "numa_id {}".format(indent,
                                   pmd_thread, pmd_thread['core_id'],
                                   pmd_thread['numa_id']))
@@ -564,7 +665,7 @@ class CmdDumpDpProvider(gdb.Command):
                   "(struct dpif_class *) 0x{:x} = {{type = {}, ...}}, "
                   "refcount = {}".
                   format(dp_class,
-                         long(dp_class['dpif_class']),
+                         int(dp_class['dpif_class']),
                          dp_class['dpif_class']['type'].string(),
                          dp_class['refcount']))
 
@@ -583,7 +684,7 @@ class CmdDumpNetdev(gdb.Command):
     @staticmethod
     def display_single_netdev(netdev, indent=0):
         indent = " " * indent
-        print("{}(struct netdev *) {}: name = {:15}, auto_classified = {:5}, "
+        print("{}(struct netdev *) {}: name = {:15}, auto_classified = {}, "
               "netdev_class = {}".
               format(indent, netdev, netdev['name'].string(),
                      netdev['auto_classified'], netdev['netdev_class']))
@@ -627,7 +728,7 @@ class CmdDumpNetdevProvider(gdb.Command):
 
         print("{}    (struct netdev_class *) 0x{:x} = {{type = {}, "
               "is_pmd = {}, ...}}, ".
-              format(indent, long(reg_class['class']),
+              format(indent, int(reg_class['class']),
                      reg_class['class']['type'].string(),
                      reg_class['class']['is_pmd']))
 
@@ -644,7 +745,7 @@ class CmdDumpNetdevProvider(gdb.Command):
 
             print("{}    (struct vport_class *) 0x{:x} = "
                   "{{ dpif_port = {}, ... }}".
-                  format(indent, long(vport), dpif_port))
+                  format(indent, int(vport), dpif_port))
 
     def invoke(self, arg, from_tty):
         netdev_classes = get_global_variable('netdev_classes')
@@ -719,10 +820,11 @@ class CmdDumpOvsList(gdb.Command):
             if typeobj is None or member is None:
                 print("(struct ovs_list *) {}".format(node))
             else:
-                print("({} *) {} =".format(
+                print("({} *) {} {}".format(
                     typeobj,
                     container_of(node,
-                                 gdb.lookup_type(typeobj).pointer(), member)))
+                                 gdb.lookup_type(typeobj).pointer(), member),
+                    "=" if dump else ""))
                 if dump:
                     print("  {}\n".format(container_of(
                         node,
@@ -734,8 +836,8 @@ class CmdDumpOvsList(gdb.Command):
 # Implements the GDB "ovs_dump_simap" command
 #
 class CmdDumpSimap(gdb.Command):
-    """Dump all nodes of an ovs_list give
-    Usage: ovs_dump_ovs_list <struct simap *>
+    """Dump all key, value entries of a simap
+    Usage: ovs_dump_simap <struct simap *>
     """
 
     def __init__(self):
@@ -756,17 +858,52 @@ class CmdDumpSimap(gdb.Command):
         values = dict()
         max_name_len = 0
         for name, value in ForEachSIMAP(simap.dereference()):
-            values[name.string()] = long(value)
+            values[name.string()] = int(value)
             if len(name.string()) > max_name_len:
                 max_name_len = len(name.string())
 
-        for name in sorted(values.iterkeys()):
+        for name in sorted(values.keys()):
             print("{}: {} / 0x{:x}".format(name.ljust(max_name_len),
                                            values[name], values[name]))
 
 
 #
-# Implements the GDB "ovs_dump_simap" command
+# Implements the GDB "ovs_dump_smap" command
+#
+class CmdDumpSmap(gdb.Command):
+    """Dump all key, value pairs of a smap
+    Usage: ovs_dump_smap <struct smap *>
+    """
+
+    def __init__(self):
+        super(CmdDumpSmap, self).__init__("ovs_dump_smap",
+                                          gdb.COMMAND_DATA)
+
+    def invoke(self, arg, from_tty):
+        arg_list = gdb.string_to_argv(arg)
+
+        if len(arg_list) != 1:
+            print("ERROR: Missing argument!\n")
+            print(self.__doc__)
+            return
+
+        smap = gdb.parse_and_eval(arg_list[0]).cast(
+            gdb.lookup_type('struct smap').pointer())
+
+        values = dict()
+        max_key_len = 0
+        for key, value in ForEachSMAP(smap.dereference()):
+            values[key.string()] = value.string()
+            if len(key.string()) > max_key_len:
+                max_key_len = len(key.string())
+
+        for key in sorted(values.keys()):
+            print("{}: {}".format(key.ljust(max_key_len),
+                                  values[key]))
+
+
+#
+# Implements the GDB "ovs_dump_udpif_keys" command
 #
 class CmdDumpUdpifKeys(gdb.Command):
     """Dump all nodes of an ovs_list give
@@ -777,6 +914,7 @@ class CmdDumpUdpifKeys(gdb.Command):
                         <udpif_name> and <udpif_address> are omitted the
                         available udpif structures are displayed.
       short           : Only dump ukey structure addresses, no content details
+      no_count        : Do not count the number of ukeys, as it might be slow
     """
 
     def __init__(self):
@@ -785,14 +923,22 @@ class CmdDumpUdpifKeys(gdb.Command):
 
     def count_all_ukeys(self, udpif):
         count = 0
+        spinner = ProgressIndicator("Counting all ukeys: ")
+
         for j in range(0, N_UMAPS):
+            spinner.update()
             count += udpif['ukeys'][j]['cmap']['impl']['p']['n']
+
+        spinner.done()
         return count
 
     def dump_all_ukeys(self, udpif, indent=0, short=False):
         indent = " " * indent
+        spinner = ProgressIndicator("Walking ukeys: ")
         for j in range(0, N_UMAPS):
+            spinner.update()
             if udpif['ukeys'][j]['cmap']['impl']['p']['n'] != 0:
+                spinner.pause()
                 print("{}(struct umap *) {}:".
                       format(indent, udpif['ukeys'][j].address))
                 for ukey in ForEachCMAP(udpif['ukeys'][j]['cmap'],
@@ -813,13 +959,13 @@ class CmdDumpUdpifKeys(gdb.Command):
                               format(
                                   indent_b, str(uuid.UUID(
                                       "{:08x}{:08x}{:08x}{:08x}".
-                                      format(long(ukey['ufid']['u32'][3]),
-                                             long(ukey['ufid']['u32'][2]),
-                                             long(ukey['ufid']['u32'][1]),
-                                             long(ukey['ufid']['u32'][0]))))))
+                                      format(int(ukey['ufid']['u32'][3]),
+                                             int(ukey['ufid']['u32'][2]),
+                                             int(ukey['ufid']['u32'][1]),
+                                             int(ukey['ufid']['u32'][0]))))))
 
                     print("{}hash = 0x{:8x}, pmd_id = {}".
-                          format(indent_b, long(ukey['hash']), ukey['pmd_id']))
+                          format(indent_b, int(ukey['hash']), ukey['pmd_id']))
                     print("{}state = {}".format(indent_b, ukey['state']))
                     print("{}n_packets = {}, n_bytes = {}".
                           format(indent_b,
@@ -828,7 +974,7 @@ class CmdDumpUdpifKeys(gdb.Command):
                     print("{}used = {}, tcp_flags = 0x{:04x}".
                           format(indent_b,
                                  ukey['stats']['used'],
-                                 long(ukey['stats']['tcp_flags'])))
+                                 int(ukey['stats']['tcp_flags'])))
 
                     #
                     # TODO: Would like to add support for dumping key, mask
@@ -841,10 +987,14 @@ class CmdDumpUdpifKeys(gdb.Command):
                     #         nlattr['nla_type'].cast(
                     #             gdb.lookup_type('enum ovs_key_attr')))
                     # print("{}key attributes = {}".format(indent_b, key))
+                spinner.resume()
+        spinner.done()
 
     def invoke(self, arg, from_tty):
         arg_list = gdb.string_to_argv(arg)
         all_udpifs = get_global_variable('all_udpifs')
+        no_count = "no_count" in arg_list
+
         if all_udpifs is None:
             return
 
@@ -852,12 +1002,15 @@ class CmdDumpUdpifKeys(gdb.Command):
         for udpif in ForEachLIST(all_udpifs, "struct udpif", "list_node"):
             udpifs[udpif['dpif']['full_name'].string()] = udpif
 
-            if len(arg_list) == 0:
+            if len(arg_list) == 0 or (
+                    len(arg_list) == 1 and arg_list[0] == "no_count"):
                 print("(struct udpif *) {}: name = {}, total keys = {}".
                       format(udpif, udpif['dpif']['full_name'].string(),
-                             self.count_all_ukeys(udpif)))
+                             self.count_all_ukeys(udpif) if not no_count
+                             else "<not counted!>"))
 
-        if len(arg_list) == 0:
+        if len(arg_list) == 0 or (
+                len(arg_list) == 1 and arg_list[0] == "no_count"):
             return
 
         if arg_list[0] in udpifs:
@@ -901,7 +1054,7 @@ class CmdShowFDB(gdb.Command):
                 gdb.lookup_type('struct ofbundle').pointer())
 
             port_name = port['name'].string()
-            port_no = long(container_of(
+            port_no = int(container_of(
                 port['ports']['next'],
                 gdb.lookup_type('struct ofport_dpif').pointer(),
                 'bundle_node')['up']['ofp_port'])
@@ -942,7 +1095,7 @@ class CmdShowFDB(gdb.Command):
             print("[(struct mac_learning *) {}]".format(ml))
 
         print("{}table.n         : {}".format(indent, ml['table']['n']))
-        print("{}secret          : 0x{:x}".format(indent, long(ml['secret'])))
+        print("{}secret          : 0x{:x}".format(indent, int(ml['secret'])))
         print("{}idle_time       : {}".format(indent, ml['idle_time']))
         print("{}max_entries     : {}".format(indent, ml['max_entries']))
         print("{}ref_count       : {}".format(indent, ml['ref_cnt']['count']))
@@ -962,9 +1115,9 @@ class CmdShowFDB(gdb.Command):
         line = "{}{:16.16}  {:-4}  {}  {:-9}".format(
             indent,
             "{}[{}]".format(port_no, port_name),
-            long(mac_entry['vlan']),
+            int(mac_entry['vlan']),
             eth_addr_to_string(mac_entry['mac']),
-            long(mac_entry['expires']))
+            int(mac_entry['expires']))
 
         if dbg:
             line += " [(struct mac_entry *) {}]".format(mac_entry)
@@ -1025,7 +1178,7 @@ class CmdShowFDB(gdb.Command):
                 max_name_len = len(node['up']['name'].string())
 
         if len(arg_list) == 0:
-            for name in sorted(all_name.iterkeys()):
+            for name in sorted(all_name.keys()):
                 print("{}: (struct mac_learning *) {}".
                       format(name.ljust(max_name_len),
                              all_name[name]['ml']))
@@ -1091,10 +1244,13 @@ class CmdShowUpcall(gdb.Command):
 
             count = 0
             j = i
+            spinner = ProgressIndicator("Counting all ukeys: ")
             while j < N_UMAPS:
+                spinner.update()
                 count += udpif['ukeys'][j]['cmap']['impl']['p']['n']
                 j += int(udpif['n_revalidators'])
 
+            spinner.done()
             print("{}  {}: (keys {}){}".
                   format(indent, revalidator['id'], count, dbg_str))
 
@@ -1111,6 +1267,46 @@ class CmdShowUpcall(gdb.Command):
             self.display_udpif_upcall(udpif, 0, "dbg" in arg_list)
 
 
+#
+# Implements the GDB "ovs_dump_ofpacts" command
+#
+class CmdDumpOfpacts(gdb.Command):
+    """Dump all actions in an ofpacts set
+    Usage: ovs_dump_ofpacts <struct ofpact *> <ofpacts_len>
+
+       <struct ofpact *> : Pointer to set of ofpact structures.
+       <ofpacts_len>     : Total length of the set.
+
+    Example dumping all actions when in the clone_xlate_actions() function:
+
+    (gdb) ovs_dump_ofpacts actions actions_len
+    (struct ofpact *) 0x561c7be487c8: {type = OFPACT_SET_FIELD, raw = 255 '', len = 24}
+    (struct ofpact *) 0x561c7be487e0: {type = OFPACT_SET_FIELD, raw = 255 '', len = 24}
+    (struct ofpact *) 0x561c7be487f8: {type = OFPACT_SET_FIELD, raw = 255 '', len = 24}
+    (struct ofpact *) 0x561c7be48810: {type = OFPACT_SET_FIELD, raw = 255 '', len = 32}
+    (struct ofpact *) 0x561c7be48830: {type = OFPACT_SET_FIELD, raw = 255 '', len = 24}
+    (struct ofpact *) 0x561c7be48848: {type = OFPACT_RESUBMIT, raw = 38 '&', len = 16}
+    """
+    def __init__(self):
+        super(CmdDumpOfpacts, self).__init__("ovs_dump_ofpacts",
+                                             gdb.COMMAND_DATA)
+
+    def invoke(self, arg, from_tty):
+        arg_list = gdb.string_to_argv(arg)
+
+        if len(arg_list) != 2:
+            print("usage: ovs_dump_ofpacts <struct ofpact *> <ofpacts_len>")
+            return
+
+        ofpacts = gdb.parse_and_eval(arg_list[0]).cast(
+            gdb.lookup_type('struct ofpact').pointer())
+
+        length = gdb.parse_and_eval(arg_list[1])
+
+        for node in ForEachOFPACTS(ofpacts, length):
+            print("(struct ofpact *) {}: {}".format(node, node.dereference()))
+
+
 #
 # Initialize all GDB commands
 #
@@ -1122,8 +1318,10 @@ CmdDumpDpNetdevPorts()
 CmdDumpDpProvider()
 CmdDumpNetdev()
 CmdDumpNetdevProvider()
+CmdDumpOfpacts()
 CmdDumpOvsList()
 CmdDumpSimap()
+CmdDumpSmap()
 CmdDumpUdpifKeys()
 CmdShowFDB()
 CmdShowUpcall()
